"""openinverter custom SDO protocol unit tests"""

import unittest

from openinverter_can_tool import constants as oi
from openinverter_can_tool.oi_node import (CanMessage, Direction, MapEntry,
                                           OpenInverterNode)
from openinverter_can_tool.paramdb import OIVariable

from .network_test_case import NetworkTestCase

TX = 1
RX = 2

# Reduce test verbosity
# pylint: disable=missing-function-docstring


class TestOpenInverterNode(NetworkTestCase):
    """
    Test the custom openinverter node protocol by example
    """

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self._node_type = OpenInverterNode

    def test_serialno(self):
        self.data = [
            (TX, b'\x40\x00\x50\x02\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x50\x02\x29\x30\x19\x87'),
            (TX, b'\x40\x00\x50\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x50\x01\x48\x86\x49\x49'),
            (TX, b'\x40\x00\x50\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x50\x00\x54\xFF\x70\x06')
        ]
        serialno = self.node.serial_no()

        # "serial" command returns:
        # 87193029:49498648:670FF54
        assert serialno == b'\x87\x19\x30\x29\x49\x49\x86\x48\x06\x70\xff\x54'

    def test_save_command(self):
        self.data = [
            (TX, b'\x23\x02\x50\x00\x00\x00\x00\x00'),
            (RX, b'\x60\x02\x50\x00\x00\x00\x00\x00')
        ]
        self.node.save()

    def test_load_command(self):
        self.data = [
            (TX, b'\x23\x02\x50\x01\x00\x00\x00\x00'),
            (RX, b'\x60\x02\x50\x01\x00\x00\x00\x00')
        ]
        self.node.load()

    def test_reset_command(self):
        self.data = [
            (TX, b'\x23\x02\x50\x02\x00\x00\x00\x00'),
            (RX, b'\x60\x02\x50\x02\x00\x00\x00\x00')
        ]
        self.node.reset()

    def test_defaults_command(self):
        self.data = [
            (TX, b'\x23\x02\x50\x03\x00\x00\x00\x00'),
            (RX, b'\x60\x02\x50\x03\x00\x00\x00\x00')
        ]
        self.node.load_defaults()

    def test_normal_start_command(self):
        self.data = [
            (TX, b'\x23\x02\x50\x04\x01\x00\x00\x00'),
            (RX, b'\x60\x02\x50\x04\x01\x00\x00\x00')
        ]
        self.node.start()

    def test_manual_start_command(self):
        self.data = [
            (TX, b'\x23\x02\x50\x04\x02\x00\x00\x00'),
            (RX, b'\x60\x02\x50\x04\x02\x00\x00\x00')
        ]
        self.node.start(mode=oi.START_MODE_MANUAL)

    def test_stop_command(self):
        self.data = [
            (TX, b'\x23\x02\x50\x05\x00\x00\x00\x00'),
            (RX, b'\x60\x02\x50\x05\x00\x00\x00\x00')
        ]
        self.node.stop()

    def test_list_empty_tx_and_rx_map(self):
        self.data = [
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x00\x00\x00\x02\x06'),
            (TX, b'\x40\x80\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x80\x31\x00\x00\x00\x02\x06')
        ]

        # The original capture covers both TX followed by RX listing of the map
        # but the API separates these two operations. Combine the requests into
        # a single test.
        tx_map = self.node.list_can_map(Direction.TX)
        assert not tx_map

        rx_map = self.node.list_can_map(Direction.RX)
        assert not rx_map

    def test_list_tx_map_single_message_and_single_param(self):
        # From a capture of the command:
        # oic can list
        # 0x101:
        # tx.0.0 param='tmphs' pos=24 length=8 gain=-1.0 offset=0
        self.data = [
            # First CAN ID
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x00\x01\x01\x00\x00'),

            # First CAN ID - first param: id, position and length
            (TX, b'\x40\x00\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x01\xE3\x07\x18\x08'),

            # First CAN ID - first param: gain and offset
            (TX, b'\x40\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x02\x18\xFC\xFF\x00'),

            # First CAN ID - second param: not present
            (TX, b'\x40\x00\x31\x03\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x03\x00\x00\x02\x06'),

            # Second CAN ID - not present
            (TX, b'\x40\x01\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x01\x31\x00\x00\x00\x02\x06')
        ]

        can_map = self.node.list_can_map(Direction.TX)

        assert len(can_map) == 1
        msg = can_map[0]

        assert msg.can_id == 0x101
        assert not msg.is_extended_frame

        assert len(msg.params) == 1
        param = msg.params[0]

        assert param.param_id == 2019
        assert param.position == 24
        assert param.length == 8
        assert param.gain == -1.0
        assert param.offset == 0

    def test_list_tx_map_single_can_id_two_params(self):
        # From a capture of the command:
        # oic can list
        # 0x101:
        # tx.0.0 param='tmphs' pos=24 length=8 gain=-1.0 offset=0
        # tx.0.1 param='tmpm' pos=0 length=8 gain=1.0 offset=0
        self.data = [
            # First CAN ID
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x00\x01\x01\x00\x00'),

            # First CAN ID - first param: id, position and length
            (TX, b'\x40\x00\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x01\xE3\x07\x18\x08'),

            # First CAN ID - first param: gain and offset
            (TX, b'\x40\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x02\x18\xFC\xFF\x00'),

            # First CAN ID - second param: id, position and length
            (TX, b'\x40\x00\x31\x03\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x03\xE4\x07\x00\x08'),

            # First CAN ID - second param: gain and offset
            (TX, b'\x40\x00\x31\x04\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x04\xE8\x03\x00\x00'),

            # First CAN ID - third param: not present
            (TX, b'\x40\x00\x31\x05\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x05\x00\x00\x02\x06'),

            # Second CAN ID - not present
            (TX, b'\x40\x01\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x01\x31\x00\x00\x00\x02\x06'),
        ]

        can_map = self.node.list_can_map(Direction.TX)

        assert len(can_map) == 1
        msg = can_map[0]

        assert msg.can_id == 0x101

        assert len(msg.params) == 2

        param = msg.params[0]
        assert param.param_id == 2019
        assert param.position == 24
        assert param.length == 8
        assert param.gain == -1.0
        assert param.offset == 0

        param = msg.params[1]
        assert param.param_id == 2020
        assert param.position == 0
        assert param.length == 8
        assert param.gain == 1.0
        assert param.offset == 0

    def test_list_tx_map_two_can_ids_single_param(self):
        # Manually synthesised CAN packets equivalent to:
        # oic can list
        # 0x001:
        # tx.0.0 param='tmphs' pos=24 length=8 gain=-1.0 offset=0
        # 0x7ff:
        # tx.1.0 param='tmpm' pos=0 length=8 gain=1.0 offset=0
        self.data = [
            # First CAN ID
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x00\x01\x00\x00\x00'),

            # First CAN ID - first param: id, position and length
            (TX, b'\x40\x00\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x01\xE3\x07\x18\x08'),

            # First CAN ID - first param: gain and offset
            (TX, b'\x40\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x02\x18\xFC\xFF\x00'),

            # First CAN ID - second param: not present
            (TX, b'\x40\x00\x31\x03\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x03\x00\x00\x02\x06'),

            # Second CAN ID
            (TX, b'\x40\x01\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x01\x31\x00\xff\x07\x00\x00'),

            # Second CAN ID - first param: id, position, length
            (TX, b'\x40\x01\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x01\x31\x01\xE4\x07\x00\x08'),

            # Second CAN ID - first param: gain and offset
            (TX, b'\x40\x01\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x43\x01\x31\x02\xE8\x03\x00\x00'),

            # Second CAN ID - second param: not present
            (TX, b'\x40\x01\x31\x03\x00\x00\x00\x00'),
            (RX, b'\x80\x01\x31\x03\x00\x00\x02\x06'),

            # Third CAN ID - not present
            (TX, b'\x40\x02\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x02\x31\x00\x00\x00\x02\x06'),
        ]

        can_map = self.node.list_can_map(Direction.TX)

        assert len(can_map) == 2
        msg = can_map[0]

        assert msg.can_id == 0x001

        assert len(msg.params) == 1
        param = msg.params[0]
        assert param.param_id == 2019
        assert param.position == 24
        assert param.length == 8
        assert param.gain == -1.0
        assert param.offset == 0

        msg = can_map[1]
        assert msg.can_id == 0x7ff

        assert len(msg.params) == 1
        param = msg.params[0]
        assert param.param_id == 2020
        assert param.position == 0
        assert param.length == 8
        assert param.gain == 1.0
        assert param.offset == 0

    def test_list_tx_map_negative_gain_offset(self):
        # From a capture of the command:
        # oic can list
        # 0x101:
        # tx.0.0 param='tmphs' pos=24 length=8 gain=-8388.608 offset=-128
        self.data = [
            # First CAN ID
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x00\x01\x01\x00\x00'),

            # First CAN ID - first param: id, position and length
            (TX, b'\x40\x00\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x01\xE3\x07\x18\x08'),

            # First CAN ID - first param: gain and offset
            (TX, b'\x40\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x02\x00\x00\x80\x80'),

            # First CAN ID - second param: not present
            (TX, b'\x40\x00\x31\x03\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x03\x00\x00\x02\x06'),

            # Second CAN ID - not present
            (TX, b'\x40\x01\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x01\x31\x00\x00\x00\x02\x06')
        ]

        can_map = self.node.list_can_map(Direction.TX)

        assert len(can_map) == 1
        msg = can_map[0]

        assert msg.can_id == 0x101

        assert len(msg.params) == 1
        param = msg.params[0]

        assert param.param_id == 2019
        assert param.position == 24
        assert param.length == 8
        assert param.gain == -8388.608
        assert param.offset == -128

    def test_list_tx_map_big_endian(self):
        # Manually synthesised CAN packets equivalent to:
        # oic can list
        # 0x103:
        # tx.0.0 param='tmpm' pos=7 length=-8 gain=1.0 offset=0
        self.data = [
            # First CAN ID
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x00\x03\x01\x00\x00'),

            # First CAN ID - first param: id, position and length
            (TX, b'\x40\x00\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x01\xE4\x07\x07\xF8'),

            # First CAN ID - first param: gain and offset
            (TX, b'\x40\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x02\xE8\x03\x00\x00'),

            # First CAN ID - second param: not present
            (TX, b'\x40\x00\x31\x03\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x03\x00\x00\x02\x06'),

            # Second CAN ID - not present
            (TX, b'\x40\x01\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x01\x31\x00\x00\x00\x02\x06')
        ]

        can_map = self.node.list_can_map(Direction.TX)

        assert len(can_map) == 1
        msg = can_map[0]

        assert msg.can_id == 0x103

        assert len(msg.params) == 1
        param = msg.params[0]

        assert param.param_id == 2020
        assert param.position == 7
        assert param.length == -8
        assert param.gain == 1.0
        assert param.offset == 0

    def test_list_tx_map_single_can_id_corrupt_param(self):
        # Manually synthesised CAN packets with the gain/offset fields not
        # provided
        self.data = [
            # First CAN ID
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x00\x01\x00\x00\x00'),

            # First CAN ID - first param: id, position and length
            (TX, b'\x40\x00\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x01\xE3\x07\x18\x08'),

            # First CAN ID - first param: gain and offset not-present
            (TX, b'\x40\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x02\x00\x00\x02\x06'),

            # Second CAN ID - not-present
            (TX, b'\x40\x01\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x01\x31\x00\x00\x00\x02\x06')
        ]

        can_map = self.node.list_can_map(Direction.TX)

        assert len(can_map) == 0

    def test_list_tx_map_single_can_id_no_params(self):
        # Manually synthesised CAN packets with no param fields at all
        self.data = [
            # First CAN ID
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x00\x01\x00\x00\x00'),

            # First CAN ID - first param: id, position and length not-present
            (TX, b'\x40\x00\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x01\x00\x00\x02\x06'),

            # Second CAN ID - not-present
            (TX, b'\x40\x01\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x01\x31\x00\x00\x00\x02\x06')
        ]

        can_map = self.node.list_can_map(Direction.TX)

        assert len(can_map) == 0

    def test_list_tx_map_single_can_id_corrupt_second_param(self):
        # Synthesised CAN frame equivalent to the command:
        # oic can list
        # 0x101:
        # tx.0.0 param='tmphs' pos=24 length=8 gain=-1.0 offset=0
        self.data = [
            # First CAN ID
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x00\x01\x01\x00\x00'),

            # First CAN ID - first param: id, position and length
            (TX, b'\x40\x00\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x01\xE3\x07\x18\x08'),

            # First CAN ID - first param: gain and offset
            (TX, b'\x40\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x02\x18\xFC\xFF\x00'),

            # First CAN ID - second param: id, position and length
            (TX, b'\x40\x00\x31\x03\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x03\xE4\x07\x00\x08'),

            # First CAN ID - second param: gain and offset - not-present
            (TX, b'\x40\x00\x31\x04\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x04\x00\x00\x02\x06'),

            # Second CAN ID - not present
            (TX, b'\x40\x01\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x01\x31\x00\x00\x00\x02\x06'),
        ]

        can_map = self.node.list_can_map(Direction.TX)

        assert len(can_map) == 1
        msg = can_map[0]

        assert msg.can_id == 0x101

        assert len(msg.params) == 1

        param = msg.params[0]
        assert param.param_id == 2019
        assert param.position == 24
        assert param.length == 8
        assert param.gain == -1.0
        assert param.offset == 0

    def test_map_transmit_parameter_successfully(self):
        # from a capture of the command:
        # oic can add tx 0x101 tmpm 0 8 1.0 0
        self.data = [
            (TX, b'\x23\x00\x30\x00\x01\x01\x00\x00'),
            (RX, b'\x60\x00\x30\x00\x01\x01\x00\x00'),
            (TX, b'\x23\x00\x30\x01\xE3\x07\x00\x08'),
            (RX, b'\x60\x00\x30\x01\xE3\x07\x00\x08'),
            (TX, b'\x23\x00\x30\x02\xE8\x03\x00\x00'),
            (RX, b'\x60\x00\x30\x02\xE8\x03\x00\x00')
        ]
        tmphs = OIVariable("tmphs", 2019)
        self.node.add_can_map_entry(
            can_id=0x101,
            direction=Direction.TX,
            param_id=tmphs.id,
            position=0,
            length=8,
            gain=1.0,
            offset=0)

    def test_map_transmit_param_with_negative_unity_gain(self):
        # from a capture of the command:
        # oic can add tx 0x101 tmphs 24 8 -1.0 0
        self.data = [
            (TX, b'\x23\x00\x30\x00\x01\x01\x00\x00'),
            (RX, b'\x60\x00\x30\x00\x01\x01\x00\x00'),
            (TX, b'\x23\x00\x30\x01\xE3\x07\x18\x08'),
            (RX, b'\x60\x00\x30\x01\xE3\x07\x18\x08'),
            (TX, b'\x23\x00\x30\x02\x18\xFC\xFF\x00'),
            (RX, b'\x60\x00\x30\x02\x18\xFC\xFF\x00')
        ]
        tmphs = OIVariable("tmphs", 2019)
        self.node.add_can_map_entry(
            can_id=0x101,
            direction=Direction.TX,
            param_id=tmphs.id,
            position=24,
            length=8,
            gain=-1.0,
            offset=0)

    def test_map_transmit_parameter_max_negative_offset(self):
        # manually synthesised can packets
        self.data = [
            # can_id request
            (TX, b'\x23\x00\x30\x00\x01\x01\x00\x00'),
            (RX, b'\x60\x00\x30\x00\x01\x01\x00\x00'),

            # param, position and length request
            (TX, b'\x23\x00\x30\x01\xE3\x07\x00\x08'),
            (RX, b'\x60\x00\x30\x01\xE3\x07\x00\x08'),

            # gain and offset request
            (TX, b'\x23\x00\x30\x02\xE8\x03\x00\x80'),
            (RX, b'\x60\x00\x30\x02\xE8\x03\x00\x80')
        ]
        tmphs = OIVariable("tmphs", 2019)
        self.node.add_can_map_entry(
            can_id=0x101,
            direction=Direction.TX,
            param_id=tmphs.id,
            position=0,
            length=8,
            gain=1.0,
            offset=-128)

    def test_map_transmit_parameter_max_gain(self):
        # manually synthesised can packets
        self.data = [
            # can_id request
            (TX, b'\x23\x00\x30\x00\x01\x01\x00\x00'),
            (RX, b'\x60\x00\x30\x00\x01\x01\x00\x00'),

            # param, position and length request
            (TX, b'\x23\x00\x30\x01\xE3\x07\x00\x08'),
            (RX, b'\x60\x00\x30\x01\xE3\x07\x00\x08'),

            # gain and offset request
            (TX, b'\x23\x00\x30\x02\xff\xff\x7f\x00'),
            (RX, b'\x60\x00\x30\x02\xff\xff\x7f\x00')
        ]
        tmphs = OIVariable("tmphs", 2019)
        self.node.add_can_map_entry(
            can_id=0x101,
            direction=Direction.TX,
            param_id=tmphs.id,
            position=0,
            length=8,
            gain=8388.607,
            offset=0)

    def test_map_transmit_parameter_max_negative_gain(self):
        # manually synthesised can packets
        self.data = [
            # can_id request
            (TX, b'\x23\x00\x30\x00\x01\x01\x00\x00'),
            (RX, b'\x60\x00\x30\x00\x01\x01\x00\x00'),

            # param, position and length request
            (TX, b'\x23\x00\x30\x01\xE3\x07\x00\x08'),
            (RX, b'\x60\x00\x30\x01\xE3\x07\x00\x08'),

            # gain and offset request
            (TX, b'\x23\x00\x30\x02\x00\x00\x80\x00'),
            (RX, b'\x60\x00\x30\x02\x00\x00\x80\x00')
        ]
        tmphs = OIVariable("tmphs", 2019)
        self.node.add_can_map_entry(
            can_id=0x101,
            direction=Direction.TX,
            param_id=tmphs.id,
            position=0,
            length=8,
            gain=-8388.608,
            offset=0)

    def test_map_receive_max_all_arguments(self):
        # manually synthesised can packets
        self.data = [
            # can_id request
            (TX, b'\x23\x01\x30\x00\xff\x07\x00\x00'),
            (RX, b'\x60\x01\x30\x00\xff\x07\x00\x00'),

            # param, position and length request
            (TX, b'\x23\x01\x30\x01\xff\x7f\x3f\x20'),
            (RX, b'\x60\x01\x30\x01\xff\x7f\x3f\x20'),

            # gain and offset request
            (TX, b'\x23\x01\x30\x02\xff\xff\x7f\x7f'),
            (RX, b'\x60\x01\x30\x02\xff\xff\x7f\x7f')
        ]
        big_param = OIVariable("fiction", 32767)
        self.node.add_can_map_entry(
            can_id=0x7ff,
            direction=Direction.RX,
            param_id=big_param.id,
            position=63,
            length=32,
            gain=8388.607,
            offset=127)

    def test_map_transmit_big_endian_successfully(self):
        # Manually synthesized equivalent to the command:
        # oic can add tx 0x101 tmpm 7 -8 1.0 0
        self.data = [
            (TX, b'\x23\x00\x30\x00\x01\x01\x00\x00'),
            (RX, b'\x60\x00\x30\x00\x01\x01\x00\x00'),
            (TX, b'\x23\x00\x30\x01\xE3\x07\x07\xF8'),
            (RX, b'\x60\x00\x30\x01\xE3\x07\x07\xF8'),
            (TX, b'\x23\x00\x30\x02\xE8\x03\x00\x00'),
            (RX, b'\x60\x00\x30\x02\xE8\x03\x00\x00')
        ]
        tmphs = OIVariable("tmphs", 2019)
        self.node.add_can_map_entry(
            can_id=0x101,
            direction=Direction.TX,
            param_id=tmphs.id,
            position=7,
            length=-8,
            gain=1.0,
            offset=0)

    def test_map_transmit_zero_can_id(self):
        # Manually synthesized packets equivalent to:
        # oic can add tx 0 tmpm 0 8 1.0 0
        self.data = [
            (TX, b'\x23\x00\x30\x00\x00\x00\x00\x00'),
            (RX, b'\x60\x00\x30\x00\x00\x00\x00\x00'),
            (TX, b'\x23\x00\x30\x01\xE3\x07\x00\x08'),
            (RX, b'\x60\x00\x30\x01\xE3\x07\x00\x08'),
            (TX, b'\x23\x00\x30\x02\xE8\x03\x00\x00'),
            (RX, b'\x60\x00\x30\x02\xE8\x03\x00\x00')
        ]
        tmphs = OIVariable("tmphs", 2019)
        self.node.add_can_map_entry(
            can_id=0,
            direction=Direction.TX,
            param_id=tmphs.id,
            position=0,
            length=8,
            gain=1.0,
            offset=0)

    def test_map_param_out_of_range_can_id(self):
        self.data = []

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x800,
                direction=Direction.TX,
                param_id=1,
                position=0,
                length=8,
                gain=1.0,
                offset=0)

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=-1,
                direction=Direction.TX,
                param_id=1,
                position=0,
                length=8,
                gain=1.0,
                offset=0)

    def test_map_param_out_of_range_direction(self):
        self.data = []

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x101,
                direction=Direction(42),
                param_id=1,
                position=0,
                length=8,
                gain=1.0,
                offset=0)

    def test_map_param_out_of_range_position(self):
        self.data = []

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x101,
                direction=Direction.TX,
                param_id=1,
                position=-1,
                length=8,
                gain=1.0,
                offset=0)

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x101,
                direction=Direction.TX,
                param_id=1,
                position=64,
                length=8,
                gain=1.0,
                offset=0)

    def test_map_param_out_of_range_length(self):
        self.data = []

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x101,
                direction=Direction.TX,
                param_id=1,
                position=0,
                length=0,
                gain=1.0,
                offset=0)

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x101,
                direction=Direction.TX,
                param_id=1,
                position=0,
                length=33,
                gain=1.0,
                offset=0)

    def test_map_param_out_of_range_gain(self):
        self.data = []

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x101,
                direction=Direction.TX,
                param_id=1,
                position=0,
                length=8,
                gain=-10000.0,
                offset=0)

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x101,
                direction=Direction.TX,
                param_id=1,
                position=0,
                length=8,
                gain=10000.0,
                offset=0)

    def test_map_param_out_of_range_offset(self):
        self.data = []

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x101,
                direction=Direction.TX,
                param_id=1,
                position=0,
                length=8,
                gain=1.0,
                offset=-129)

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x101,
                direction=Direction.TX,
                param_id=1,
                position=0,
                length=8,
                gain=1.0,
                offset=128)

    def test_remove_first_mapped_param(self):
        # from a capture of the command:
        # oic can remove tx.0.0
        self.data = [
            (TX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x23\x00\x31\x02\x00\x00\x00\x00')
        ]
        assert self.node.remove_can_map_entry(Direction.TX, 0, 0)

    def test_remove_fourth_param_from_second_can_messsage(self):
        # from a capture of the command:
        # oic can remove tx.1.3
        self.data = [
            (TX, b'\x23\x01\x31\x08\x00\x00\x00\x00'),
            (RX, b'\x23\x01\x31\x08\x00\x00\x00\x00')
        ]
        assert self.node.remove_can_map_entry(Direction.TX, 1, 3)

    def test_remove_not_present_rx_mapping(self):
        # from a capture of the command:
        # oic can remove rx.5.5
        # with no RX map defined
        self.data = [
            (TX, b'\x23\x85\x31\x0C\x00\x00\x00\x00'),
            (RX, b'\x80\x85\x31\x0C\x00\x00\x02\x06')
        ]
        assert not self.node.remove_can_map_entry(Direction.RX, 5, 5)

    def test_clear_map_tx_no_mappings_present(self):
        # From a capture of running:
        # oic can remove rx.0.0
        self.data = [
            (TX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x02\x00\x00\x02\x06')
        ]
        assert not self.node.clear_map(Direction.TX)

    def test_clear_map_tx_large_map(self):
        # From a capture of running:
        # oic can remove tx.0.0
        # until it reports "Unable to find CAN map entry."
        self.data = [
            (TX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (TX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (TX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (TX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (TX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (TX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (TX, b'\x23\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x02\x00\x00\x02\x06')
        ]
        assert not self.node.clear_map(Direction.TX)

    def test_clear_map_rx_single_message_single_param_map(self):
        # From a capture of running:
        # oic can remove rx.0.0
        # until it reports "Unable to find CAN map entry."
        self.data = [
            (TX, b'\x23\x80\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x23\x80\x31\x02\x00\x00\x00\x00'),
            (TX, b'\x23\x80\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x80\x80\x31\x02\x00\x00\x02\x06')
        ]
        assert not self.node.clear_map(Direction.RX)

    def test_add_multiple_messages_in_a_single_map(self):
        # Captured from running the command sequence:
        # oic can add tx 0x101 tmpm 0 8 little 1.0 0
        # oic can add tx 0x102 tmphs 32 32 little 2.0 0
        self.data = [
            (TX, b'\x23\x00\x30\x00\x01\x01\x00\x00'),
            (RX, b'\x60\x00\x30\x00\x01\x01\x00\x00'),
            (TX, b'\x23\x00\x30\x01\xE4\x07\x00\x08'),
            (RX, b'\x60\x00\x30\x01\xE4\x07\x00\x08'),
            (TX, b'\x23\x00\x30\x02\xE8\x03\x00\x00'),
            (RX, b'\x60\x00\x30\x02\xE8\x03\x00\x00'),
            (TX, b'\x23\x00\x30\x00\x02\x01\x00\x00'),
            (RX, b'\x60\x00\x30\x00\x02\x01\x00\x00'),
            (TX, b'\x23\x00\x30\x01\xE3\x07\x20\x20'),
            (RX, b'\x60\x00\x30\x01\xE3\x07\x20\x20'),
            (TX, b'\x23\x00\x30\x02\xD0\x07\x00\x00'),
            (RX, b'\x60\x00\x30\x02\xD0\x07\x00\x00')
        ]

        tmpm = OIVariable("tmpm", 2020)
        tmphs = OIVariable("tmphs", 2019)

        msg_map = [
            CanMessage(
                can_id=0x101,
                params=[MapEntry(tmpm.id, 0, 8,  1.0, 0)]
            ),
            CanMessage(
                can_id=0x102,
                params=[MapEntry(tmphs.id, 32, 32, 2.0, 0)]
            )
        ]

        self.node.add_can_map(Direction.TX, msg_map)


if __name__ == "__main__":
    unittest.main()



"""Useful constants"""

# CANopen SDO indexes used by openinverter systems

# Index for unique serial numbers
SERIALNO_INDEX = 0x5000

# Parameter database checksum
PARAM_DB_CHECKSUM_SUBINDEX = 3

# Index for long segmented strings
STRINGS_INDEX = 0x5001

# CANopen SDO subindex for JSON parameter databases. Used with STRINGS_INDEX
PARAM_DB_SUBINDEX = 0

# Index for sending commands - somewhat abusing the SDO protocol
COMMAND_INDEX = 0x5002

SAVE_COMMAND_SUBINDEX = 0
LOAD_COMMAND_SUBINDEX = 1
RESET_COMMAND_SUBINDEX = 2
DEFAULTS_COMMAND_SUBINDEX = 3
START_COMMAND_SUBINDEX = 4
STOP_COMMAND_SUBINDEX = 5

# CAN mapping indexes
CAN_MAP_TX_INDEX = 0x3000
CAN_MAP_RX_INDEX = 0x3001
CAN_MAP_LIST_TX_INDEX = 0x3100
CAN_MAP_LIST_RX_INDEX = 0x3180

MAP_CAN_ID_SUBINDEX = 0
MAP_PARAM_POS_LEN_SUBINDEX = 1
MAP_GAIN_OFFSET_SUBINDEX = 2

MAP_EXTENDED_FRAME_FLAG = 0x20000000
MAP_EXTENDED_FRAME_MASK = 0x1fffffff

# Start modes - tracks the _modes enum in stm32-sine
START_MODE_NORMAL = 1
START_MODE_MANUAL = 2
START_MODE_BOOST = 3
START_MODE_BUCK = 4
START_MODE_SINE = 5
START_MODE_ACHEAT = 6

# SDO abort error codes - Not defined by canopen for some reason
SDO_ABORT_OBJECT_NOT_AVAILABLE = 0x06020000

# Names for common directories
APPNAME = "openinverter_can_tool"
APPAUTHOR = "openinverter"



"""
Unit test parameter database functions
"""
import filecmp
import json
import unittest
from pathlib import Path
from typing import cast

import canopen.objectdictionary
import pytest

from openinverter_can_tool.fpfloat import fixed_from_float
from openinverter_can_tool.paramdb import (OIVariable,
                                           import_cached_database,
                                           import_database,
                                           import_database_json,
                                           import_remote_database)

from .oi_sim import OISimulatedNode

TEST_DATA_DIR = Path(__file__).parent / "test_data" / "paramdb"

# Reduce test verbosity
# pylint: disable=missing-function-docstring


class OpeninverterVariable(unittest.TestCase):
    """
    Unit test the OIVariable class used to represent the not quite CANopen
    variable representation method used by openinverter
    """

    def test_zero_id(self):
        """ openinverter parameters all start with an index of 0x2100"""
        var = OIVariable("zero_id",  0)
        self.assertEqual(var.index, 0x2100)
        self.assertEqual(var.subindex, 0)

    def test_small_id(self):
        """ Check that ids < 0xff only affect the subindex byte"""
        var = OIVariable("id",  1)
        self.assertEqual(var.index, 0x2100)
        self.assertEqual(var.subindex, 1)

        var = OIVariable("id",  42)
        self.assertEqual(var.index, 0x2100)
        self.assertEqual(var.subindex, 42)

        var = OIVariable("id",  0xff)
        self.assertEqual(var.index, 0x2100)
        self.assertEqual(var.subindex, 0xff)

    def test_large_id(self):
        """ Check that larger IDs are split into the correct index/subindex"""
        var = OIVariable("large_id", 2015)
        self.assertEqual(var.index, 0x2107)
        self.assertEqual(var.subindex, 0xdf)

    def test_return_id(self):
        """ Check that the openinverter ID is stored as well as the CANopen
        index and sub-index. """
        var = OIVariable("id",  2015)
        self.assertEqual(var.id, 2015)
        self.assertEqual(var.index, 0x2107)
        self.assertEqual(var.subindex, 0xdf)

    def test_modify_index(self):
        """ Check that it is possible to modify the sub-index and have this
        reflected in the id returned """
        var = OIVariable("id",  2015)
        var.index = 0x2100
        var.subindex = 0xff
        self.assertEqual(var.id, 0xff)
        self.assertEqual(var.index, 0x2100)
        self.assertEqual(var.subindex, 0xff)


class DatabaseImport(unittest.TestCase):
    """
    Unit test the JSON parameter database import functionality
    """

    def test_invalid_db_filename(self):
        """Verify that a garbage filename fails with an exception"""
        with pytest.raises(FileNotFoundError):
            import_database(Path("not_a_real_file.json"))

    def test_zero_byte_db_file(self):
        """Verify that a zero length file raises a JSON parse exception"""
        with pytest.raises(json.decoder.JSONDecodeError):
            import_database(TEST_DATA_DIR / "zero-bytes.json")

    def test_empty_db_file(self):
        """Verify that an empty file loads but contains no entries"""
        self.assertCountEqual(import_database(
            TEST_DATA_DIR / "empty-but-valid.json"), [])

    def test_single_param(self):
        """Verify that a simple database with a single parameter
        loads correctly"""
        database = import_database(TEST_DATA_DIR / "single-param.json")
        assert database["param1"]
        item = cast(OIVariable, database["param1"])
        self.assertEqual(item.index, 0x2100)
        self.assertEqual(item.subindex, 1)
        self.assertEqual(item.unit, "km / h")
        self.assertEqual(item.min, fixed_from_float(0))
        self.assertEqual(item.max, fixed_from_float(100))
        self.assertEqual(item.default, fixed_from_float(5))
        self.assertEqual(item.factor, 32)
        self.assertEqual(item.data_type, canopen.objectdictionary.INTEGER32)
        self.assertTrue(item.isparam)
        self.assertEqual(item.category, "Category")

    def test_complex_params(self):
        """Verify that a more complex database with a variety of parameters
        and some values loads correctly"""
        database = import_database(TEST_DATA_DIR / "complex.json")

        expected_params = [
            {"name": "curkp", "isparam": True, "unit": "",
             "min": 0, "max": 20000, "default": 32,
             "category": "Params",
             "index": 0x2100, "subindex": 107},
            {"name": "dirmode", "isparam": True,
             "unit": "0=Button, 1=Switch, 2=ButtonReversed, 3=SwitchReversed, "
                     "4=DefaultForward",
             "min": 0, "max": 4, "default": 1,
             "category": "Params",
             "index": 0x2100, "subindex": 95},
            {"name": "potmin", "isparam": True, "unit": "dig",
             "min": 0, "max": 4095, "default": 0,
             "category": "Throttle",
             "index": 0x2100, "subindex": 17},
            {"name": "potmax", "isparam": True, "unit": "dig",
             "min": 0, "max": 4095, "default": 4095,
             "category": "Throttle",
             "index": 0x2100, "subindex": 18},
            {"name": "cpuload", "isparam": False, "unit": "%",
             "index": 0x2107, "subindex": 0xF3}
        ]

        # Basic size check
        self.assertEqual(len(database.names), len(expected_params))

        # verify each of the exepected params exist
        for param in expected_params:
            item = cast(OIVariable, database[param["name"]])
            self.assertEqual(item.index, param["index"])
            self.assertEqual(item.subindex, param["subindex"])
            self.assertEqual(item.unit, param["unit"])
            self.assertEqual(item.isparam, param["isparam"])

            # optional fields only present for params not values
            if item.isparam:
                self.assertEqual(item.min, fixed_from_float(param["min"]))
                self.assertEqual(item.max, fixed_from_float(param["max"]))
                self.assertEqual(
                    item.default, fixed_from_float(param["default"]))
                self.assertEqual(item.category, param["category"])
            else:
                self.assertEqual(item.min, None)
                self.assertEqual(item.max, None)
                self.assertEqual(item.default, None)
                self.assertEqual(item.category, None)

            self.assertEqual(item.factor, 32)
            self.assertEqual(
                item.data_type, canopen.objectdictionary.INTEGER32)

    def test_unicode_param(self):
        """Verify that databases with Unicode work. We need this for degree
        symbols at least but emojis are just as fun."""
        database = import_database(TEST_DATA_DIR / "unicode.json")
        assert database["param1"]
        item = cast(OIVariable, database["param1"])
        self.assertEqual(item.index, 0x2100)
        self.assertEqual(item.subindex, 1)
        self.assertEqual(item.unit, "°")
        self.assertEqual(item.min, fixed_from_float(0))
        self.assertEqual(item.max, fixed_from_float(100))
        self.assertEqual(item.default, fixed_from_float(5))
        self.assertEqual(item.factor, 32)
        self.assertEqual(item.data_type, canopen.objectdictionary.INTEGER32)
        self.assertTrue(item.isparam)
        self.assertEqual(item.category, "😬")

    def test_raw_json_dict(self):
        """Verify that it is possible to parse a raw JSON dictionary without
        requiring a file. Include extraneous attributes on parameters that are
        present on stm32-sine from 5.25-R onwards. Also include a parameter
        without any ID which should be ignored."""

        raw_json = {
            "curkp": {"unit": "", "minimum": "0", "maximum": "20000",
                      "default": "32", "isparam": True, "category": "Motor",
                      "id": "107", "i": 0},
            "dirmode": {"unit": "0=Button, 1=Switch, 2=ButtonReversed, "
                        "3=SwitchReversed, 4=DefaultForward", "id": 95,
                        "value": 1.00, "isparam": True, "minimum": 0.00,
                        "maximum": 4.00, "default": 1.00,
                        "category": "Motor", "i": 15},
            "serial": {"unit": "", "value": "87193029", "isparam": False}
        }

        database = import_database_json(raw_json)

        expected_params = [
            {"name": "curkp", "isparam": True, "unit": "",
             "min": 0, "max": 20000, "default": 32,
             "category": "Motor",
             "index": 0x2100, "subindex": 107},
            {"name": "dirmode", "isparam": True,
             "unit": "0=Button, 1=Switch, 2=ButtonReversed, 3=SwitchReversed, "
                     "4=DefaultForward",
             "min": 0, "max": 4, "default": 1,
             "category": "Motor",
             "index": 0x2100, "subindex": 95}
        ]

        # Basic size check
        self.assertEqual(len(database.names), len(expected_params))

        # verify each of the expected params exist
        for param in expected_params:
            item = cast(OIVariable, database[param["name"]])
            self.assertEqual(item.index, param["index"])
            self.assertEqual(item.subindex, param["subindex"])
            self.assertEqual(item.unit, param["unit"])
            self.assertEqual(item.isparam, param["isparam"])

            # optional fields only present for params not values
            if item.isparam:
                self.assertEqual(item.min, fixed_from_float(param["min"]))
                self.assertEqual(item.max, fixed_from_float(param["max"]))
                self.assertEqual(
                    item.default, fixed_from_float(param["default"]))
                self.assertEqual(item.category, param["category"])
            else:
                self.assertEqual(item.min, None)
                self.assertEqual(item.max, None)
                self.assertEqual(item.default, None)
                self.assertEqual(item.category, None)

            self.assertEqual(item.factor, 32)
            self.assertEqual(item.data_type,
                             canopen.objectdictionary.INTEGER32)

    def test_remote_db(self):
        """Verify that it is possible to load a database located on a remote
        CAN bus node."""

        simulator = OISimulatedNode(13)
        simulator.LoadDatabase(TEST_DATA_DIR / "complex.json")

        database = import_remote_database(simulator.network, 13)

        expected_params = [
            {"name": "curkp", "isparam": True, "unit": "",
             "min": 0, "max": 20000, "default": 32,
             "category": "Params",
             "index": 0x2100, "subindex": 107},
            {"name": "dirmode", "isparam": True,
             "unit": "0=Button, 1=Switch, 2=ButtonReversed, 3=SwitchReversed, "
                     "4=DefaultForward",
             "min": 0, "max": 4, "default": 1,
             "category": "Params",
             "index": 0x2100, "subindex": 95},
            {"name": "potmin", "isparam": True, "unit": "dig",
             "min": 0, "max": 4095, "default": 0,
             "category": "Throttle",
             "index": 0x2100, "subindex": 17},
            {"name": "potmax", "isparam": True, "unit": "dig",
             "min": 0, "max": 4095, "default": 4095,
             "category": "Throttle",
             "index": 0x2100, "subindex": 18},
            {"name": "cpuload", "isparam": False, "unit": "%",
             "index": 0x2107, "subindex": 0xF3}
        ]

        # Basic size check
        self.assertEqual(len(database.names), len(expected_params))

        # verify each of the expected params exist
        for param in expected_params:
            item = cast(OIVariable, database[param["name"]])
            self.assertEqual(item.index, param["index"])
            self.assertEqual(item.subindex, param["subindex"])
            self.assertEqual(item.unit, param["unit"])
            self.assertEqual(item.isparam, param["isparam"])

            # optional fields only present for params not values
            if item.isparam:
                self.assertEqual(item.min, fixed_from_float(param["min"]))
                self.assertEqual(item.max, fixed_from_float(param["max"]))
                self.assertEqual(
                    item.default, fixed_from_float(param["default"]))
                self.assertEqual(item.category, param["category"])
            else:
                self.assertEqual(item.min, None)
                self.assertEqual(item.max, None)
                self.assertEqual(item.default, None)
                self.assertEqual(item.category, None)

            self.assertEqual(item.factor, 32)
            self.assertEqual(
                item.data_type, canopen.objectdictionary.INTEGER32)

    def test_remote_db_with_zero_bytes(self):
        """Due to a race condition in openinverter firmware the database can
        contain additional 0x00 bytes interspersed with the expected byte
        stream. Verify that these databases can be loaded correctly from a
        remote node."""

        simulator = OISimulatedNode(13)
        simulator.LoadDatabase(
            TEST_DATA_DIR / "complex-with-added-zero-bytes.json")

        database = import_remote_database(simulator.network, 13)

        expected_params = [
            {"name": "curkp", "isparam": True, "unit": "",
             "min": 0, "max": 20000, "default": 32,
             "category": "Params",
             "index": 0x2100, "subindex": 107},
            {"name": "dirmode", "isparam": True,
             "unit": "0=Button, 1=Switch, 2=ButtonReversed, 3=SwitchReversed, "
                     "4=DefaultForward",
             "min": 0, "max": 4, "default": 1,
             "category": "Params",
             "index": 0x2100, "subindex": 95},
            {"name": "potmin", "isparam": True, "unit": "dig",
             "min": 0, "max": 4095, "default": 0,
             "category": "Throttle",
             "index": 0x2100, "subindex": 17},
            {"name": "potmax", "isparam": True, "unit": "dig",
             "min": 0, "max": 4095, "default": 4095,
             "category": "Throttle",
             "index": 0x2100, "subindex": 18},
            {"name": "cpuload", "isparam": False, "unit": "%",
             "index": 0x2107, "subindex": 0xF3}
        ]

        # Basic size check
        self.assertEqual(len(database.names), len(expected_params))

        # verify each of the expected params exist
        for param in expected_params:
            item = cast(OIVariable, database[param["name"]])
            self.assertEqual(item.index, param["index"])
            self.assertEqual(item.subindex, param["subindex"])
            self.assertEqual(item.unit, param["unit"])
            self.assertEqual(item.isparam, param["isparam"])

            # optional fields only present for params not values
            if item.isparam:
                self.assertEqual(item.min, fixed_from_float(param["min"]))
                self.assertEqual(item.max, fixed_from_float(param["max"]))
                self.assertEqual(
                    item.default, fixed_from_float(param["default"]))
                self.assertEqual(item.category, param["category"])
            else:
                self.assertEqual(item.min, None)
                self.assertEqual(item.max, None)
                self.assertEqual(item.default, None)
                self.assertEqual(item.category, None)

            self.assertEqual(item.factor, 32)
            self.assertEqual(
                item.data_type, canopen.objectdictionary.INTEGER32)

    def test_remote_unicode_db_with_zero_bytes(self):
        """Due to a race condition in openinverter firmware the database can
        contain additional NUL or 0x00 bytes. Verify that a databases with
        unicode utf-8 sequences with extra zero bytes can be loaded correctly
        from a remote node."""

        simulator = OISimulatedNode(13)
        simulator.LoadDatabase(
            TEST_DATA_DIR / "unicode-with-added-zero-bytes.json")

        database = import_remote_database(simulator.network, 13)

        assert database["param1"]
        item = cast(OIVariable, database["param1"])
        self.assertEqual(item.index, 0x2100)
        self.assertEqual(item.subindex, 1)
        self.assertEqual(item.unit, "°")
        self.assertEqual(item.min, fixed_from_float(0))
        self.assertEqual(item.max, fixed_from_float(100))
        self.assertEqual(item.default, fixed_from_float(5))
        self.assertEqual(item.factor, 32)
        self.assertEqual(item.data_type, canopen.objectdictionary.INTEGER32)
        self.assertTrue(item.isparam)
        self.assertEqual(item.category, "😬")

    def test_enum_dict(self):
        """Provide a dictionary with a variety of enumeration parameters.
        Verify that these are correctly parsed.
        """

        raw_json = {
            "dirmode": {"unit": "0=Button, 1=Switch, 2=ButtonReversed, "
                        "3=SwitchReversed, 4=DefaultForward", "id": 95,
                        "value": 1.00, "isparam": True, "minimum": 0.00,
                        "maximum": 4.00, "default": 1.00,
                        "category": "Motor", "i": 15},
            "snsm": {"unit": "12=KTY83-110, 13=KTY84-130, 14=Leaf, "
                     "15=KTY81-110, 16=Toyota, 21=OutlanderFront, "
                     "22=EpcosB57861-S, 23=ToyotaGen2", "minimum": "12",
                     "maximum": "23", "default": "12", "isparam": True,
                     "category": "Motor", "id": "46"},
            "dir": {"unit": "-1=Reverse, 0=Neutral, 1=Forward",
                    "isparam": False, "id": "2018"},
            "din_mprot": {"unit": "0=Error, 1=Ok, 2=na", "isparam": False,
                          "id": "2026"},

            # Note trailing comma
            "lasterr": {"unit": "0=NONE, 1=OVERCURRENT, 2=THROTTLE1, "
                        "3=THROTTLE2, 4=CANTIMEOUT, 5=EMCYSTOP, 6=MPROT, "
                        "7=DESAT, 8=OVERVOLTAGE, 9=ENCODER, 10=PRECHARGE, "
                        "11=TMPHSMAX, 12=CURRENTLIMIT, 13=PWMSTUCK, "
                        "14=HICUROFS1, 15=HICUROFS2, 16=HIRESOFS, "
                        "17=LORESAMP, 18=TMPMMAX,",
                        "isparam": False, "id": "2038"},

            "version": {"unit": "4=5.24.R-foc", "isparam": False, "id": "2039"}
        }

        database = import_database_json(raw_json)

        expected_params = [
            {"name": "dirmode",
             "enums": {0: "Button", 1: "Switch", 2: "ButtonReversed",
                       3: "SwitchReversed", 4: "DefaultForward"}},
            {"name": "snsm",
             "enums": {12: "KTY83-110", 13: "KTY84-130", 14: "Leaf",
                       15: "KTY81-110", 16: "Toyota", 21: "OutlanderFront",
                       22: "EpcosB57861-S", 23: "ToyotaGen2"}},
            {"name": "dir",
             "enums": {-1: "Reverse", 0: "Neutral", 1: "Forward"}},
            {"name": "din_mprot",
             "enums": {0: "Error", 1: "Ok", 2: "na"}},
            {"name": "lasterr",
             "enums": {0: "NONE", 1: "OVERCURRENT", 2: "THROTTLE1",
                       3: "THROTTLE2", 4: "CANTIMEOUT", 5: "EMCYSTOP",
                       6: "MPROT", 7: "DESAT", 8: "OVERVOLTAGE", 9: "ENCODER",
                       10: "PRECHARGE", 11: "TMPHSMAX", 12: "CURRENTLIMIT",
                       13: "PWMSTUCK", 14: "HICUROFS1", 15: "HICUROFS2",
                       16: "HIRESOFS", 17: "LORESAMP", 18: "TMPMMAX"}},
            {"name": "version",
             "enums": {4: "5.24.R-foc"}},
        ]

        # Basic size check
        self.assertEqual(len(database.names), len(expected_params))

        # verify each of the expected params exist
        for param in expected_params:
            item = database[param["name"]]
            assert isinstance(item, OIVariable)
            self.assertFalse(item.bit_definitions)

            expected_enums = param["enums"]

            self.assertEqual(len(item.value_descriptions), len(expected_enums))
            for value, description in expected_enums.items():
                self.assertEqual(
                    item.value_descriptions[value], description)

    def test_bitfield_dict(self):
        """Provide a dictionary with a variety of bitfield parameters.
        Verify that these are correctly parsed.
        """

        raw_json = {
            "canio": {"unit": "1=Cruise, 2=Start, 4=Brake, 8=Fwd, 16=Rev, "
                      "32=Bms",
                      "isparam": False, "id": "2022"},
            "status": {"unit": "0=None, 1=UdcLow, 2=UdcHigh, 4=UdcBelowUdcSw, "
                       "8=UdcLim, 16=EmcyStop, 32=MProt, 64=PotPressed, "
                       "128=TmpHs, 256=WaitStart", "isparam": False,
                       "id": "2044"}
        }

        database = import_database_json(raw_json)

        expected_params = [
            {"name": "canio",
             "bitfield": {1: "Cruise", 2: "Start", 4: "Brake", 8: "Fwd",
                          16: "Rev", 32: "Bms"}},
            {"name": "status",
             "bitfield": {0: "None", 1: "UdcLow", 2: "UdcHigh",
                          4: "UdcBelowUdcSw", 8: "UdcLim", 16: "EmcyStop",
                          32: "MProt", 64: "PotPressed", 128: "TmpHs",
                          256: "WaitStart"}},
        ]

        # Basic size check
        self.assertEqual(len(database.names), len(expected_params))

        # verify each of the expected params exist
        for param in expected_params:
            item = database[param["name"]]
            assert isinstance(item, OIVariable)
            self.assertFalse(item.value_descriptions)

            expected_bitfield = param["bitfield"]

            self.assertEqual(len(item.bit_definitions),
                             len(expected_bitfield))
            for value, description in expected_bitfield.items():
                self.assertEqual(
                    item.bit_definitions[value], description)

    def test_badly_punctuated_enum_missing_comma(self):
        """Extracted from issue #4 a badly punctuated enum should try
        to fix up the list in the same way as esp8266-web-interface
        """

        raw_json = {
            "Inverter": {
                # Note the lack of a comma between 7 and 8
                "unit": "0=None, 1=Leaf_Gen1, 2=GS450H, 3=UserCAN, 4=OpenI, "
                "5=Prius_Gen3, 6=Outlander, 7=GS300H 8=RearOutlander",
                "id": 5,
                "value": 0.00,
                "isparam": True,
                "minimum": 0.00,
                "maximum": 8.00,
                "default": 0.00,
                "category": "General Setup",
                "i": 0
            }}

        database = import_database_json(raw_json)

        expected_enums = {
            0: "None", 1: "Leaf_Gen1", 2: "GS450H", 3: "UserCAN", 4: "OpenI",
            5: "Prius_Gen3", 6: "Outlander", 7: "GS300H",  8: "RearOutlander"
        }

        assert len(database) == 1
        item = database["Inverter"]
        self.assertEqual(len(item.value_descriptions), len(expected_enums))
        for value, description in expected_enums.items():
            self.assertEqual(
                item.value_descriptions[value], description)

    def test_badly_punctuated_enum_full_stop_rather_than_comma(self):
        """Extracted from issue #4 a badly punctuated enum should try
        to fix up the list in the same way as esp8266-web-interface
        """

        raw_json = {
            "CAN3Speed": {
                "unit": "0=k33.3, 1=k500. 2=k100",
                "id": 77,
                "value": 0.00,
                "isparam": True,
                "minimum": 0.00,
                "maximum": 2.00,
                "default": 0.00,
                "category": "Communication",
                "i": 53
            }}

        database = import_database_json(raw_json)

        expected_enums = {
            0: "k33.3", 1: "k500.", 2: "k100"
        }

        assert len(database) == 1
        item = database["CAN3Speed"]

        self.assertEqual(len(item.value_descriptions), len(expected_enums))
        for value, description in expected_enums.items():
            self.assertEqual(
                item.value_descriptions[value], description)

    def test_badly_punctuated_enum_with_no_spaces(self):
        """Extracted from issue #4 a badly punctuated enum without any spaces
        should still be parsed fine
        """

        raw_json = {
            "Out1Func": {
                "unit": "0=None, 1=ChaDeMoAlw, 2=OBCEnable, 3=HeaterEnable, "
                "4=RunIndication, 5=WarnIndication,6=CoolantPump, "
                "7=NegContactor, 8=BrakeLight, 9=ReverseLight, 10=HeatReq, "
                "11=HVRequest,12=DCFCRequest, 13=BrakeVacPump, 14=PwmTim3",
                "id": 80,
                "value": 6.00,
                "isparam": True,
                "minimum": 0.00,
                "maximum": 13.00,
                "default": 6.00,
                "category": "General Purpose I/O",
                "i": 87
            }
        }

        database = import_database_json(raw_json)

        assert len(database) == 1
        item = database["Out1Func"]
        assert item.value_descriptions[11] == "HVRequest"
        assert item.value_descriptions[12] == "DCFCRequest"
        assert item.value_descriptions[13] == "BrakeVacPump"
        assert item.value_descriptions[14] == "PwmTim3"
        assert len(item.value_descriptions) == 15

    def test_badly_punctuated_enum_with_no_value_name(self):
        """Verify that a massively poorly formatted enum fails gracefully"""

        raw_json = {
            "Option": {
                "unit": "0=starts-ok, 1, 2=ends-well",
                "id": 77,
                "value": 0.00,
                "isparam": True,
                "minimum": 0.00,
                "maximum": 2.00,
                "default": 0.00,
                "category": "Communication",
                "i": 53
            }}

        database = import_database_json(raw_json)

        assert len(database) == 1
        item = database["Option"]
        assert item.unit == ("0=starts-ok, 1, 2=ends-well [DB FORMAT ERROR]")
        assert len(item.value_descriptions) == 0


class TestCachedDatabases:
    """
    Unit test caching of JSON parameter databases
    """

    def test_new_empty_cache_location(self, tmp_path: Path):
        simulator = OISimulatedNode(42)
        simulator.checksum = 12345678
        simulator.LoadDatabase(TEST_DATA_DIR / "single-param.json")

        cache = tmp_path / "empty-but-non-existent"
        assert not cache.is_dir()

        database = import_cached_database(simulator.network, 42, cache)

        assert cache.is_dir()

        assert database["param1"]

        cached_file = next(cache.iterdir(), None)
        assert cached_file
        assert cached_file.is_file()
        assert cached_file.stat().st_size > 0

    def test_long_new_cache_path(self, tmp_path: Path):
        simulator = OISimulatedNode(42)
        simulator.checksum = 12345678
        simulator.LoadDatabase(TEST_DATA_DIR / "single-param.json")

        cache = tmp_path / "a" / "deep" / "new" / "path"
        assert not cache.is_dir()

        database = import_cached_database(simulator.network, 42, cache)

        assert cache.is_dir()

        assert database["param1"]

    def test_empty_but_present_cache_location(self, tmp_path: Path):
        simulator = OISimulatedNode(42)
        simulator.checksum = 12345678
        simulator.LoadDatabase(TEST_DATA_DIR / "single-param.json")

        cache = tmp_path / "empty-but-exists"
        cache.mkdir()

        assert len(list(cache.iterdir())) == 0

        database = import_cached_database(simulator.network, 42, cache)

        assert database["param1"]
        item = cast(OIVariable, database["param1"])
        assert item.index == 0x2100
        assert item.subindex == 1
        assert item.unit == "km / h"
        assert item.min == fixed_from_float(0)
        assert item.max == fixed_from_float(100)
        assert item.default == fixed_from_float(5)
        assert item.factor == 32
        assert item.data_type == canopen.objectdictionary.INTEGER32
        assert item.isparam
        assert item.category == "Category"

        cached_file = next(cache.iterdir(), None)
        assert cached_file
        assert cached_file.is_file()
        assert cached_file.stat().st_size > 0

    def test_database_is_cached(self, tmp_path: Path):
        simulator = OISimulatedNode(42)
        simulator.checksum = 12345678
        simulator.LoadDatabase(TEST_DATA_DIR / "single-param.json")

        cache = tmp_path

        # prime the cache
        database = import_cached_database(simulator.network, 42, cache)

        assert database["param1"]

        # Load a completely different database but don't update the checksum
        simulator.LoadDatabase(TEST_DATA_DIR / "complex.json")

        # Load the database again which should load from the cache
        database = import_cached_database(simulator.network, 42, cache)

        # verify we still have the single parameter
        assert database["param1"]
        item = cast(OIVariable, database["param1"])
        assert item.index == 0x2100
        assert item.subindex == 1
        assert item.unit == "km / h"
        assert item.min == fixed_from_float(0)
        assert item.max == fixed_from_float(100)
        assert item.default == fixed_from_float(5)
        assert item.factor == 32
        assert item.data_type == canopen.objectdictionary.INTEGER32
        assert item.isparam
        assert item.category == "Category"

        # There should just be a single file in the cache
        cached_file = next(cache.iterdir(), None)
        assert cached_file
        assert cached_file.is_file()
        assert cached_file.stat().st_size > 0

        # verify that a parameters from the file we loaded to the remote node
        # is not present
        assert "curkp" not in database
        assert "dirmode" not in database
        assert "potmin" not in database
        assert "potmax" not in database
        assert "cpuload" not in database

    def test_cached_database_is_updated(self, tmp_path: Path):
        simulator = OISimulatedNode(42)
        simulator.checksum = 12345678
        simulator.LoadDatabase(TEST_DATA_DIR / "single-param.json")

        cache = tmp_path

        # prime the cache
        database = import_cached_database(simulator.network, 42, cache)

        assert database["param1"]

        # Load a completely different database and update the checksum on the
        # remote node
        simulator.LoadDatabase(TEST_DATA_DIR / "complex.json")
        simulator.checksum = 4567890

        # Load the database again which should update from the remote node
        database = import_cached_database(simulator.network, 42, cache)

        # verify we have parameters from the new database
        assert database["curkp"]
        assert database["dirmode"]
        assert database["potmin"]
        assert database["potmax"]
        assert database["cpuload"]

        # verify that nothing remains of the old parameters
        assert "param1" not in database

        # There should two cached files for each version of the database
        assert len(list(cache.iterdir())) == 2

    def test_multiple_nodes_generate_multiple_cached_databases(
            self,
            tmp_path: Path):
        cache = tmp_path

        # Set up up the first node
        simulator = OISimulatedNode(42)
        simulator.checksum = 12345678
        simulator.LoadDatabase(TEST_DATA_DIR / "single-param.json")

        # Load the database from the first node
        database = import_cached_database(simulator.network, 42, cache)

        assert database["param1"]

        # Set up a second node with the same database
        simulator = OISimulatedNode(99)
        simulator.checksum = 12345678
        simulator.LoadDatabase(TEST_DATA_DIR / "single-param.json")

        # Load the database from the second node
        database = import_cached_database(simulator.network, 99, cache)

        assert database["param1"]

        # There should two cached files for each node and they should be
        # identical
        cache_files = list(cache.iterdir())
        assert len(cache_files) == 2
        assert filecmp.cmp(cache_files[0], cache_files[1], shallow=False)


if __name__ == '__main__':
    unittest.main()



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
import pathlib
import re
from glob import glob
from setuptools import setup, find_packages

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

# Strip the build status as this only makes sense on github
long_description = re.sub(r'\[!\[Build status.*\)\n\n', '', long_description)

setup(
    name="openinverter-can-tool",
    version="0.3.2",
    description="Tool to configure and operate openinverter systems over CAN",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/davefiddes/openinverter-can-tool",
    author="David J. Fiddes",
    author_email="D.J@fiddes.net",
    license="MIT License",

    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3 :: Only",
    ],

    keywords="openinverter, canopen",

    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.8, <4",

    install_requires=[
        "click",
        "canopen",
        "appdirs",
        "python-can[serial]",
        "cantools"],

    extras_require={
        "dev": [
            "check-manifest",
            "flake8",
            "pre-commit"
        ],
        "test": [
            "coverage",
            "pytest",
            "approvaltests",
            "pytest-approvaltests",
            "pytest-cov"
        ],
    },

    # No data files are expected within the package
    package_data={},

    # Pull in all our example parameter databases and documentation
    data_files=[("parameter-databases",
                 glob("parameter-databases/*.json")),
                ("docs",
                 glob("docs/*.md") + glob("docs/*.png"))],

    # The main command-line tool
    entry_points={
        "console_scripts": [
            "oic=openinverter_can_tool.__main__:cli"
        ],
    },

    project_urls={
        "Bug Reports":
            "https://github.com/davefiddes/openinverter-can-tool/issues",
        "Source":
            "https://github.com/davefiddes/openinverter-can-tool",
    },
)



"""
openinverter custom SDO protocol unit tests
Extended frame specific tests
"""

import unittest

from openinverter_can_tool.oi_node import (CanMessage, Direction, MapEntry,
                                           OpenInverterNode)
from openinverter_can_tool.paramdb import OIVariable

from .network_test_case import NetworkTestCase

TX = 1
RX = 2

# Reduce test verbosity
# pylint: disable=missing-function-docstring


class TestOpenInverterNodeExtendedFrames(NetworkTestCase):
    """
    Test the custom openinverter node protocol by example
    """

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self._node_type = OpenInverterNode

    def test_map_receive_extended_can_id(self):
        # Manually synthesized packets equivalent to:
        # oic can add rx 0x312 ChargeCurrent 31 -16 0.100000001 0 --extended
        # From a question relating to integrating a boostech bms v3
        # https://openinverter.org/forum/viewtopic.php?p=72768#p72768
        self.data = [
            (TX, b'\x23\x01\x30\x00\x12\x03\x00\x20'),
            (RX, b'\x60\x01\x30\x00\x12\x03\x00\x20'),
            (TX, b'\x23\x01\x30\x01\x04\x00\x1F\xF0'),
            (RX, b'\x60\x01\x30\x01\x04\x00\x1F\xF0'),
            (TX, b'\x23\x01\x30\x02\x64\x00\x00\x00'),
            (RX, b'\x60\x01\x30\x02\x64\x00\x00\x00')
        ]
        charge_current = OIVariable("ChargeCurrent", 4)
        self.node.add_can_map_entry(
            can_id=0x312,
            direction=Direction.RX,
            param_id=charge_current.id,
            position=31,
            length=-16,
            gain=0.100000001,
            offset=0,
            is_extended_frame=True)

    def test_map_transmit_extended_can_id(self):
        # Manually synthesized packets equivalent to:
        # oic can add tx 0x12345678 tmpm 0 8 1.0 0 --extended
        self.data = [
            (TX, b'\x23\x00\x30\x00\x78\x56\x34\x32'),
            (RX, b'\x60\x00\x30\x00\x00\x00\x00\x00'),
            (TX, b'\x23\x00\x30\x01\xE3\x07\x00\x08'),
            (RX, b'\x60\x00\x30\x01\xE3\x07\x00\x08'),
            (TX, b'\x23\x00\x30\x02\xE8\x03\x00\x00'),
            (RX, b'\x60\x00\x30\x02\xE8\x03\x00\x00')
        ]
        tmphs = OIVariable("tmphs", 2019)
        self.node.add_can_map_entry(
            can_id=0x12345678,
            direction=Direction.TX,
            param_id=tmphs.id,
            position=0,
            length=8,
            gain=1.0,
            offset=0,
            is_extended_frame=True)

    def test_map_param_out_of_range_extended_can_id(self):
        self.data = []

        with self.assertRaises(ValueError):
            self.node.add_can_map_entry(
                can_id=0x20000000,
                direction=Direction.TX,
                param_id=1,
                position=0,
                length=8,
                gain=1.0,
                offset=0,
                is_extended_frame=True)

    def test_list_tx_map_single_extended_can_id_message_and_single_param(self):
        # Synthesised CAN frame equivalent to the command:
        # oic can list
        # 0x12345678:
        # tx.0.0 param='tmphs' pos=24 length=8 gain=-1.0 offset=0
        self.data = [
            # First CAN ID
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x00\x78\x56\x34\x32'),

            # First CAN ID - first param: id, position and length
            (TX, b'\x40\x00\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x01\xE3\x07\x18\x08'),

            # First CAN ID - first param: gain and offset
            (TX, b'\x40\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x02\x18\xFC\xFF\x00'),

            # First CAN ID - second param: not present
            (TX, b'\x40\x00\x31\x03\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x03\x00\x00\x02\x06'),

            # Second CAN ID - not present
            (TX, b'\x40\x01\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x01\x31\x00\x00\x00\x02\x06')
        ]

        can_map = self.node.list_can_map(Direction.TX)

        assert len(can_map) == 1
        msg = can_map[0]

        assert msg.can_id == 0x12345678
        assert msg.is_extended_frame
        assert isinstance(msg.is_extended_frame, bool)

        assert len(msg.params) == 1
        param = msg.params[0]

        assert param.param_id == 2019
        assert param.position == 24
        assert param.length == 8
        assert param.gain == -1.0
        assert param.offset == 0

    def test_list_tx_map_single_extended_can_id_with_short_id(self):
        # Synthesised CAN frame equivalent to the command:
        # oic can list
        # 0x00000101:
        # tx.0.0 param='tmphs' pos=24 length=8 gain=-1.0 offset=0
        self.data = [
            # First CAN ID
            (TX, b'\x40\x00\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x00\x01\x01\x00\x20'),

            # First CAN ID - first param: id, position and length
            (TX, b'\x40\x00\x31\x01\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x01\xE3\x07\x18\x08'),

            # First CAN ID - first param: gain and offset
            (TX, b'\x40\x00\x31\x02\x00\x00\x00\x00'),
            (RX, b'\x43\x00\x31\x02\x18\xFC\xFF\x00'),

            # First CAN ID - second param: not present
            (TX, b'\x40\x00\x31\x03\x00\x00\x00\x00'),
            (RX, b'\x80\x00\x31\x03\x00\x00\x02\x06'),

            # Second CAN ID - not present
            (TX, b'\x40\x01\x31\x00\x00\x00\x00\x00'),
            (RX, b'\x80\x01\x31\x00\x00\x00\x02\x06')
        ]

        can_map = self.node.list_can_map(Direction.TX)

        assert len(can_map) == 1
        msg = can_map[0]

        assert msg.can_id == 0x101
        assert msg.is_extended_frame
        assert isinstance(msg.is_extended_frame, bool)

        assert len(msg.params) == 1
        param = msg.params[0]

        assert param.param_id == 2019
        assert param.position == 24
        assert param.length == 8
        assert param.gain == -1.0
        assert param.offset == 0

    def test_add_mixing_extended_and_standard_messages_in_a_single_map(self):
        # Captured from running the command sequence:
        # oic can add tx 0x101 tmpm 0 8 little 1.0 0
        # oic can add tx 0x102 tmphs 32 32 little 2.0 0 --extended
        self.data = [
            (TX, b'\x23\x00\x30\x00\x01\x01\x00\x00'),
            (RX, b'\x60\x00\x30\x00\x01\x01\x00\x00'),
            (TX, b'\x23\x00\x30\x01\xE4\x07\x00\x08'),
            (RX, b'\x60\x00\x30\x01\xE4\x07\x00\x08'),
            (TX, b'\x23\x00\x30\x02\xE8\x03\x00\x00'),
            (RX, b'\x60\x00\x30\x02\xE8\x03\x00\x00'),
            (TX, b'\x23\x00\x30\x00\x02\x01\x00\x20'),
            (RX, b'\x60\x00\x30\x00\x02\x01\x00\x20'),
            (TX, b'\x23\x00\x30\x01\xE3\x07\x20\x20'),
            (RX, b'\x60\x00\x30\x01\xE3\x07\x20\x20'),
            (TX, b'\x23\x00\x30\x02\xD0\x07\x00\x00'),
            (RX, b'\x60\x00\x30\x02\xD0\x07\x00\x00')
        ]

        tmpm = OIVariable("tmpm", 2020)
        tmphs = OIVariable("tmphs", 2019)

        msg_map = [
            CanMessage(
                can_id=0x101,
                params=[MapEntry(tmpm.id, 0, 8,  1.0, 0)],
                is_extended_frame=False
            ),
            CanMessage(
                can_id=0x102,
                params=[MapEntry(tmphs.id, 32, 32, 2.0, 0)],
                is_extended_frame=True
            )
        ]

        self.node.add_can_map(Direction.TX, msg_map)


if __name__ == "__main__":
    unittest.main()
